from qiskit import QuantumCircuit

BELL_STATES = ("00+11", "00-11", "01+10", "01-10")


def bell_pair(state: str = "00+11") -> QuantumCircuit:
    r"""Return circuit that produces a Bell state.
    If the argument `state` isn't passed in, the default "00+11" string is used.

    :param state: The state string describing a Bell state \(default is "00+11"\).
            Valid choices are ``"00+11", "00-11", "01+10", "01-10"``, which correspond to
            :math:`|\Phi^+\rangle, |\Phi^-\rangle, |\Psi^+\rangle` and :math:`|\Psi^-\rangle` Bell states respectively
    :raises ValueError: If state is not one of the 4 valid strings
    :return: Circuit producing a Bell state
    """
    if state not in BELL_STATES:
        raise ValueError('input state must be "00+11", "00-11", "01+10" or "01-10"')
    n = BELL_STATES.index(state)
    bell_state_qc = QuantumCircuit(2)
    if n & 1:
        bell_state_qc.x(0)
    if n & 2:
        bell_state_qc.x(1)
    bell_state_qc.h(0)
    bell_state_qc.cx(0, 1)
    if n == 3:
        bell_state_qc.swap(0, 1)
    return bell_state_qc



from __future__ import annotations

import uuid
from dataclasses import dataclass
from typing import TYPE_CHECKING

from qiskit.providers import BackendV2, Options

import bluequbit
from bluequbit.qiskit_provider.job import BlueQubitQiskitJob

if TYPE_CHECKING:
    from qiskit import QuantumCircuit
    from qiskit.result import Result


@dataclass
class BQDataBin:
    meas: Result


class BlueQubitBackendV2(BackendV2):
    def __init__(
        self,
        api_token: str | None = None,
        execution_mode="cloud",
    ):
        super().__init__()
        self.bq_client = bluequbit.init(
            api_token=api_token, execution_mode=execution_mode
        )
        self.execution_mode = execution_mode

    @classmethod
    def _default_options(cls):
        return Options(
            shots=None,
            device="cpu",
            job_name=None,
            pauli_sum=None,
            asynchronous=True,
            options=None,
        )

    @property
    def max_circuits(self):
        return None

    @property
    def target(self):
        return None

    def run(
        self,
        circuits: QuantumCircuit | list[QuantumCircuit],
        **run_options,
    ):
        if run_options.get("asynchronous") is False:
            raise ValueError(
                "Non-asynchronous execution is not supported in BlueQubitBackendV2 for Qiskit"
            )
        options = self._default_options().__dict__ | run_options

        job_id = str(uuid.uuid4())
        bq_job = BlueQubitQiskitJob(
            self,
            job_id=job_id,
            fn=self._execute,
            circuits=circuits,
            run_options=options,
        )
        bq_job.submit()
        return bq_job

    def _execute(self, circuits: QuantumCircuit | list[QuantumCircuit], options):
        jobs = self.bq_client.run(
            circuits,
            **options,
        )
        return jobs



JOB_NO_RESULT_TERMINAL_STATES = {
    "FAILED_VALIDATION",
    "CANCELED",
    "TERMINATED",
    "NOT_ENOUGH_FUNDS",
    "JOBS_LIMIT_EXCEEDED",
}
JOB_RESULTS_READY_STATES = {"COMPLETED"}
JOB_TERMINAL_STATES = JOB_RESULTS_READY_STATES.union(JOB_NO_RESULT_TERMINAL_STATES)
JOB_NON_TERMINAL_STATES = {"PENDING", "QUEUED", "RUNNING"}
JOB_STATES = JOB_NON_TERMINAL_STATES.union(JOB_TERMINAL_STATES)

DEVICE_TYPES = {
    "cpu",
    "gpu",
    "quantum",
    "tensor-network",
    "pennylane.cpu",
    "pennylane.gpu",
    "pennylane.cpu.adjoint",
    "pennylane.gpu.adjoint",
    "mps.cpu",  # Dec 2024
    "mps.gpu",
    "pauli-path",  # Dec 2024
}

MAXIMUM_NUMBER_OF_BATCH_JOBS = 500

QUEUED_CPU_JOBS_LIMIT = 5

MAXIMUM_NUMBER_OF_JOBS_FOR_RUN = 50

MAX_QUBITS_WITH_STATEVEC = 16

MAXIMUM_NUMBER_OF_SHOTS = dict.fromkeys(DEVICE_TYPES, 100_000)
MAXIMUM_NUMBER_OF_SHOTS["mps.cpu"] = 1_000_000
MAXIMUM_NUMBER_OF_SHOTS["mps.gpu"] = 1_000_000

MAX_MPS_CPU_BOND_DIMENSION = 2048
MAX_MPS_GPU_BOND_DIMENSION = 4000
MAX_TWO_QUBIT_GATE_COUNT = {"mps.cpu": 10_000, "pauli-path": 3000}

# smallest truncation threshold that is currently supported for pauli-path jobs
MIN_PAULI_PATH_TRUNCATION_THRESHOLD = 1e-5

# Maximum size in bytes for serialized circuit data
MAXIMUM_SERIALIZED_CIRCUIT_SIZE = 10_000_000  # 10 MB
MAX_JOB_TAGS_SIZE = 1_000_000  # 1 MB

MAX_JOB_NAME_LENGTH = 200



import json
import logging
import os
import time
import uuid
from collections import defaultdict
from pathlib import Path

from .http_utils import request_retriable
from .version import __version__

logger = logging.getLogger("bluequbit-python-sdk")


class BackendConnection:
    def __init__(self, api_token=None):
        super().__init__()
        config_dir = Path.home() / ".config" / "bluequbit"
        config_location = config_dir / "config.json"

        main_endpoint_from_local_env = None
        ssl_verify_from_local_env = None
        token_from_env_variable = os.environ.get("BLUEQUBIT_API_TOKEN")

        if config_location.is_file():
            with config_location.open(encoding="utf-8") as f:
                config = json.load(f)
            main_endpoint_from_local_env = config.get("main_endpoint")
            ssl_verify_from_local_env = config.get("ssl_verify")
        main_endpoint_from_local_env = os.environ.get(
            "BLUEQUBIT_MAIN_ENDPOINT", main_endpoint_from_local_env
        )
        if api_token is None:
            api_token = token_from_env_variable

        api_config = {
            "token": api_token,
            "main_endpoint": (
                "https://app.bluequbit.io/api/v1"
                if main_endpoint_from_local_env is None
                or main_endpoint_from_local_env == ""
                else main_endpoint_from_local_env
            ),
            "ssl_verify": (
                True if ssl_verify_from_local_env is None else ssl_verify_from_local_env
            ),
        }

        self._token = api_token

        self._default_headers = {
            "Authorization": f"SDK {self._token}",
            "Connection": "close",
            "User-Agent": f"BlueQubit SDK {__version__}",
        }

        self._main_endpoint = api_config["main_endpoint"]
        if self._main_endpoint != "https://app.bluequbit.io/api/v1":
            logger.warning("Using custom endpoint %s", self._main_endpoint)
        self._verify = True
        if "ssl_verify" in api_config:
            self._verify = api_config["ssl_verify"]
        self._session = None
        self._num_requests = defaultdict(int)
        self._last_request_time = time.monotonic()
        self.min_request_interval = 0.0
        self.send_request(
            req_type="GET",
            path="/jobs",
            params={"limit": 1},
        )

    def send_request(
        self,
        req_type,
        path,
        params=None,
        data=None,
        json_req=None,
        headers=None,
    ):
        url = self._main_endpoint + path

        if params is not None:
            for key, value in params.items():
                if isinstance(value, str):
                    params[key] = value.replace("\\", "\\\\")
                if isinstance(value, list):
                    params[key] = ",".join(value)

        if headers is None:
            headers_to_send = self._default_headers
        else:
            headers_to_send = dict(self._default_headers, **headers)
        request_id = str(uuid.uuid4())
        headers_to_send["X-Request-ID"] = request_id
        self._num_requests[req_type] += 1
        request_time = time.monotonic()
        time.sleep(
            max(
                0.0,
                self.min_request_interval - (request_time - self._last_request_time),
            )
        )
        self._last_request_time = request_time
        return request_retriable(
            method=req_type,
            url=url,
            data=data,
            json=json_req,
            params=params,
            headers=headers_to_send,
        )



import base64
import io
import pickle
import zlib


def is_a_qiskit_circuit(qc):
    return any(
        str(cls) == "<class 'qiskit.circuit.quantumcircuit.QuantumCircuit'>"
        for cls in type(qc).mro()
    )


def is_a_pennylane_circuit(qc):
    return str(type(qc)) == "<class 'dict'>" and any(
        str(cls) == "<class 'pennylane.tape.qscript.QuantumScript'>"
        for cls in type(qc.get("pennylane_circuit")).mro()
    )


def cirq_to_qasm2(circuit):
    import cirq
    from cirq import ops

    qubit_order = ops.QubitOrder.DEFAULT
    qubits = ops.QubitOrder.as_qubit_order(qubit_order).order_for(circuit.all_qubits())
    serialized = cirq.QasmOutput(
        operations=circuit.all_operations(),
        qubits=qubits,
        header="Generated from Cirq via BlueQubit",
        precision=10,
        version="2.0",
    )
    return str(serialized)


def encode_circuit(qc, use_cirq_qasm=False):
    """
    Encode a quantum circuit object from various quantum libraries to a
    portable format.
    """
    # We use the string representation of the object class so that we don't
    # have to require users to install Qiskit, Cirq, and Braket all at once.
    if isinstance(qc, dict) and "circuit_type" in qc:
        return qc
    circuit_type = str(type(qc))
    if circuit_type == "<class 'cirq.circuits.circuit.Circuit'>":
        if use_cirq_qasm:
            # We serialize to OQ2 string
            # https://quantumai.google/reference/python/cirq/QasmOutput
            serialized = cirq_to_qasm2(qc)
            return {"circuit_type": "Cirq_OQ2", "circuit": serialized}

        # We use Cirq-specific serialization method.
        import cirq

        # https://quantumai.google/cirq/build/interop
        serialized = cirq.to_json(qc)
        return {"circuit_type": "Cirq", "circuit": serialized}
    if circuit_type == "<class 'braket.circuits.circuit.Circuit'>":
        from braket.circuits.serialization import IRType

        serialized = qc.to_ir(IRType.OPENQASM).source
        return {"circuit_type": "Braket", "circuit": serialized}
    if is_a_qiskit_circuit(qc):
        from importlib.metadata import version

        from qiskit import qpy

        qiskit_version = version("qiskit")

        # https://qiskit.org/documentation/apidoc/qpy.html
        with io.BytesIO() as f:
            if qiskit_version[0] == "2":
                qpy.dump(qc, f, version=13)
            else:
                qpy.dump(qc, f)
            # Rewind back to the beginning of the "file".
            f.seek(0)
            serialized = f.read()
            # Convert to base64
            serialized = base64.b64encode(serialized)
            # Convert bytes to str
            serialized = serialized.decode("utf-8")
        return {"circuit_type": "Qiskit", "circuit": serialized}
    if is_a_pennylane_circuit(qc):
        return {
            "circuit_type": "Pennylane",
            "circuit": base64.b64encode(zlib.compress(pickle.dumps(qc))).decode(
                "utf-8"
            ),
        }
    raise TypeError("Unsupported circuit type", qc)


def encode_circuit_with_fallback(qc):
    # Temporary workaround: see https://trello.com/c/0SMvJuLV/266-productofsums-issue
    encoded = encode_circuit(qc)
    if encoded["circuit_type"] == "Cirq" and "ProductOfSums" in encoded["circuit"]:
        return encode_circuit(qc, use_cirq_qasm=True)
    return encoded


def decode_circuit(data):
    """
    Decode the output of encode_circuit.
    """
    if "circuit_type" not in data:
        raise Exception("Wrong data structure. 'circuit_type' must be present.")
    if "circuit" not in data:
        raise Exception("Wrong data structure. 'circuit' must be present.")
    circuit_type = data["circuit_type"]
    if circuit_type == "Cirq":
        import cirq

        return cirq.read_json(json_text=data["circuit"])
    if circuit_type == "Cirq_OQ2":
        from cirq.contrib.qasm_import import circuit_from_qasm

        return circuit_from_qasm(data["circuit"])
    if circuit_type == "Qiskit":
        from qiskit import qpy

        _bytes = str.encode(data["circuit"])
        _bytes = base64.b64decode(_bytes)
        with io.BytesIO(_bytes) as f:
            circuits = qpy.load(f)
            if len(circuits) != 1:
                raise Exception(
                    f"Unexpected number of circuits: {len(circuits)}. Must be 1."
                )
            return circuits[0]
    if circuit_type == "Pennylane":
        encoded_circuit = data["circuit"]
        decoded_bytes = base64.b64decode(encoded_circuit)
        circuit = pickle.loads(zlib.decompress(decoded_bytes))  # noqa: S301
        if isinstance(circuit, dict):
            pennylane_circuit = circuit.get("pennylane_circuit")
        else:
            raise TypeError(
                "the circuit must be dictionary containing pennylane_circuit"
            )
        return pennylane_circuit
    if circuit_type == "Braket":
        raise Exception("Braket decoding is not yet supported.")
    raise Exception("Unsupported circuit type", circuit_type)



import logging
import time

import requests

from .exceptions import BQAPIError, BQUnauthorizedAccessError

logger = logging.getLogger("bluequbit-python-sdk")

DEFAULT_TIMEOUT = (15.0, 300.0)
NUMBER_OF_RETRIES = 5


def exponential_backoff(retry_num, backoff_factor=1.0):
    return backoff_factor * (2**retry_num - 1)


def request_retriable(method, url, **kwargs):
    retry_num = 0
    resp = requests.Response()

    def _send():
        nonlocal retry_num
        nonlocal method
        nonlocal kwargs
        nonlocal resp
        if retry_num > 0:
            time.sleep(exponential_backoff(retry_num - 1))
        retry_num += 1
        try:
            resp = requests.request(method, url, timeout=DEFAULT_TIMEOUT, **kwargs)
            if resp.ok:
                return resp
            logger.debug(
                f"HTTP request from BlueQubit SDK error. Request {method} {url}. Response {resp.status_code} {resp.text}"
            )
            if retry_num <= NUMBER_OF_RETRIES and resp.status_code in [
                429,
                502,
                503,
                504,
            ]:
                logger.warning(
                    "Retrying. Num retries: %s. HTTP status code: %s.",
                    retry_num,
                    resp.status_code,
                )
                return _send()
            if resp.status_code == 401:
                raise BQUnauthorizedAccessError(
                    kwargs.get("headers", {}).get("X-Request-ID", "")
                )
            resp.raise_for_status()
        except requests.ConnectTimeout as e:
            if retry_num <= NUMBER_OF_RETRIES:
                logger.warning(
                    "Retrying. Num retries %s. ConnectTimeout: %s", retry_num, e
                )
                return _send()
            raise
        except requests.RequestException as e:
            if retry_num <= NUMBER_OF_RETRIES and method in [
                "HEAD",
                "GET",
                "PUT",
                "DELETE",
                "OPTIONS",
                "TRACE",
            ]:
                logger.warning(
                    "Retrying. Num retries %s. RequestException: %s.", retry_num, e
                )
                return _send()
            raise

    try:
        return _send()
    except Exception as e:
        if isinstance(e, BQUnauthorizedAccessError):
            raise
        raise BQAPIError(
            resp.status_code,
            resp.text,
            kwargs.get("headers", {}).get("X-Request-ID", ""),
        ) from e



# Copyright 2023 Bluequbit Inc.
#
# All Rights Reserved
#
# NOTICE: All information contained herein is, and remains the property of Bluequbit Inc.
# and its suppliers, if any. The intellectual and technical concepts contained herein are
# proprietary to Bluequbit Inc. and its suppliers and may be covered by U.S. and Foreign Patents,
# patents in process, and are protected by trade secret or copyright law.
# Dissemination of this information or reproduction of this material is strictly forbidden unless
# prior written permission is obtained from Bluequbit Inc.
#
# This file is part of a proprietary software product and is not to be copied, modified,
# redistributed, or used in any way without the express written permission of Bluequbit Inc.


import pennylane as qml
from pennylane import numpy as np
from scipy.sparse import diags
from tqdm.notebook import trange

complex_dtype = np.complex128
real_dtype = np.float64


class AdamOptim:
    def __init__(self, eta=0.03, beta1=0.9, beta2=0.999, epsilon=1e-8):
        super().__init__()
        self.m, self.v = 0, 0
        self.beta1 = beta1
        self.beta2 = beta2
        self.epsilon = epsilon
        self.eta = eta
        self.t = 1

    def step(self, grad):
        ## dw, db are from current minibatch
        ## momentum beta 1
        self.m = self.beta1 * self.m + (1 - self.beta1) * grad
        ## rms beta 2
        self.v = self.beta2 * self.v + (1 - self.beta2) * (grad**2)
        ## bias correction
        m_corr = self.m / (1 - self.beta1**self.t)
        v_corr = self.v / (1 - self.beta2**self.t)
        self.t += 1
        ## update weights and biases
        return self.eta * (m_corr / (np.sqrt(v_corr) + self.epsilon))


def _get_circuit_info(circuit):
    # todo: assert circuit is qnode
    params_dummy = np.zeros(1000000, requires_grad=True)
    specs = qml.specs(circuit)(params_dummy)
    num_params = specs[
        "num_trainable_params"
    ]  # one way to get number of params from a circuit
    num_wires = specs["resources"].num_wires
    return num_wires, num_params


def load_data(
    target_probs: np.ndarray,
    circuit_fn,  #: a QNode-able function (if it's QNode, it will work but very slow)
    num_wires: int,
    seed: int = 0,
    loss_type: str = "kl-divergence",
    num_epochs: int = 1000,
    params_init=None,
    eta=0.03,
):
    r"""Run QCBM training to find an approximation to the target_distribution"""
    if loss_type != "kl-divergence":
        print(
            f"WARNING: Got unsupported loss_type {loss_type}: falling back to kl-divergence"
        )
        loss_type = "kl-divergence"
    dev_gpu = qml.device("lightning.gpu", wires=num_wires, c_dtype=complex_dtype)
    circuit = qml.QNode(circuit_fn, dev_gpu)
    num_wires, num_params = _get_circuit_info(circuit)
    circuit_adj = qml.QNode(
        circuit_fn, dev_gpu, diff_method="adjoint", device_vjp=False
    )

    def adjoint_kl_native(params):
        wires = range(num_wires)
        ket = circuit(params).astype(complex_dtype)
        obs = -target_probs / (ket * ket.conj())
        hmat = diags(obs, format="csr")
        h = qml.SparseHamiltonian(hmat, wires)
        grad = qml.grad(circuit_adj)(params, h)
        loss = np.sum(target_probs * np.log(target_probs / (np.abs(ket) ** 2)))
        return grad, loss, ket

    np.random.seed(seed)
    best_x, best_loss = None, np.inf
    if params_init is None:
        x = np.random.rand(num_params, requires_grad=True).astype(real_dtype)
    else:
        x = np.zeros(num_params, requires_grad=True).astype(real_dtype)
        x[: len(params_init)] = params_init
    print(f"starting training with qubits: {num_wires} params: {num_params}")
    print(qml.specs(circuit)(x)["resources"])
    #     print(qml.draw(circuit_gpu)(x))
    opt = AdamOptim(eta=eta)
    break_points: dict[int, float] = {}
    # break_points = {100: 1.22, 200: 1.04, 300: 0.98, 400: 0.95, 600: 0.9,
    #                 900: 0.8, 1200 : 0.75, 1500: 0.7}
    loss = np.inf
    for epoch in trange(num_epochs):
        if epoch in break_points and loss > break_points[epoch]:
            print(f"CUTTING SHORT...{loss} > {break_points[epoch]} ")
            break
        grad, loss, pred_state = adjoint_kl_native(x)
        grad = np.array(grad).astype(real_dtype)
        if loss < best_loss:
            best_loss = loss
            if best_loss < 8 and epoch % 1 == 0:
                pred_state = circuit(x)
                l1_loss = np.sum(np.abs(target_probs - np.abs(pred_state) ** 2))
                print(
                    f"NEW* best_loss: {best_loss}: qubits: {num_wires} L1_loss: {l1_loss}"
                )
            best_x = x.copy()
        x -= opt.step(grad)
        print(f"epoch: {epoch} \t loss: {loss} \t", flush=True)
    return best_x, best_loss



from qiskit.quantum_info import Pauli


def construct_pauli_from_idx_lists(
    idx_lists: list[list[int]], num_qubits: int
) -> Pauli:
    """Construct a Pauli operator from a list of Pauli X, Y, Z indices.
    idx_lists[0] is the list of indices for the X Pauli
    idx_lists[1] is the list of indices for the Y Pauli
    idx_lists[2] is the list of indices for the Z Pauli.
    The Pauli operator is the product of the Pauli operators at the indices in the lists.
    The Pauli operator is returned as a qiskit.quantum_info.Pauli object.

    :param idx_lists: list of lists of indices
    :type idx_lists: list[list[int]]
    :param num_qubits: number of qubits
    :type num_qubits: int

    :return: Pauli operator
    :rtype: qiskit.quantum_info.Pauli
    """
    paulis = ["I"] * num_qubits
    set_x = set(idx_lists[0])
    set_y = set(idx_lists[1])
    set_z = set(idx_lists[2])
    if len(set_x | set_y | set_z) != len(set_x) + len(set_y) + len(set_z):
        raise ValueError("idx_lists must not have overlapping indices")
    for i in set_x:
        paulis[i] = "X"
    for i in set_y:
        paulis[i] = "Y"
    for i in set_z:
        paulis[i] = "Z"
    return Pauli("".join(paulis))



from __future__ import annotations

import logging
from typing import TYPE_CHECKING

import numpy as np
from qiskit.primitives import (
    BaseEstimatorV2,
    PrimitiveJob,
    PrimitiveResult,
    PubResult,
)
from qiskit.primitives.containers import DataBin, EstimatorPubLike
from qiskit.primitives.containers.estimator_pub import EstimatorPub

from .provider import BlueQubitBackendV2

if TYPE_CHECKING:
    from collections.abc import Iterable

logger = logging.getLogger("bluequbit-python-sdk")


class BlueQubitEstimatorV2(BaseEstimatorV2):
    """
    Evaluates expectation values for provided quantum circuit and observable combinations.

    Qiskit Estimator V2 interface to BlueQubit cloud computation: https://docs.quantum.ibm.com/api/qiskit-ibm-runtime/qiskit_ibm_runtime.EstimatorV2
    Based on:
    - https://github.com/Qiskit/qiskit-ibm-runtime/blob/main/qiskit_ibm_runtime/estimator.py
    - https://github.com/Qiskit/qiskit-aer/blob/main/qiskit_aer/primitives/estimator_v2.py
    """

    def __init__(
        self,
        *,
        backend: BlueQubitBackendV2 | None = None,
        options: dict | None = None,
    ):
        self._options = options if options is not None else {}
        if backend is None:
            backend = BlueQubitBackendV2(**self._options.get("backend_options", {}))
        self._run_options = self._options.get("run_options", {})
        self._backend = backend
        self._warned_precision = False

    @property
    def options(self):
        """Return the options"""
        return self._options

    def run(
        self, pubs: Iterable[EstimatorPubLike], *, precision: float | None = None
    ) -> PrimitiveJob[PrimitiveResult[PubResult]]:
        if precision is not None and not self._warned_precision:
            logger.warning(
                "BlueQubit only supports float32 precision. The precision argument is going to be ignored."
            )
            self._warned_precision = True
        coerced_pubs = [EstimatorPub.coerce(pub, precision) for pub in pubs]
        job = PrimitiveJob(self._run, coerced_pubs)
        job._submit()  # noqa: SLF001
        return job

    def _run(self, pubs: list[EstimatorPub]) -> PrimitiveResult[PubResult]:
        return PrimitiveResult([self._run_pub(pub) for pub in pubs])

    def _run_pub(self, pub: EstimatorPub) -> PubResult:
        circuit = pub.circuit
        observables = pub.observables
        parameter_values = pub.parameter_values

        bound_circuits = parameter_values.bind_all(circuit)
        bc_circuits, bc_obs = np.broadcast_arrays(bound_circuits, observables)

        evs = np.zeros_like(bc_circuits, dtype=np.float64)

        for index in np.ndindex(*bc_circuits.shape):
            bound_circuit = bc_circuits[index]
            observable = bc_obs[index]
            precision = pub.precision

            pauli_sums = list(observable.items())
            job = self._backend.run(
                bound_circuit,
                pauli_sum=pauli_sums,  # type: ignore[arg-type]
                **self._run_options,
            )
            result = job.result()
            expectation_value = np.real_if_close(result.expectation_value)
            if precision != 0 and not np.isreal(expectation_value):
                raise ValueError(
                    "Given operator is not Hermitian and noise cannot be added."
                )
            evs[index] = expectation_value

        data = DataBin(evs=evs, shape=evs.shape)
        return PubResult(
            data,
            metadata={
                "target_precision": precision,
                "circuit_metadata": pub.circuit.metadata,
            },
        )



from __future__ import annotations

from .backend import BlueQubitBackendV2


class BlueQubitProvider:
    """A Qiskit provider for accessing BlueQubit backend.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in the environment variable BLUEQUBIT_API_TOKEN.
    """

    def __init__(self, api_token: str | None = None):
        super().__init__()
        self.api_token = api_token

    def get_backend(self, execution_mode: str | None = None):
        """
        :param execution_mode: Execution mode. Can be ``"cloud"``, ``"local"`` or ``None``.
                           If ``None``, the environment variable
                           BLUEQUBIT_EXECUTION_MODE will be used if set, otherwise
                           ``"cloud"`` will be used.
        """
        return BlueQubitBackendV2(
            api_token=self.api_token, execution_mode=execution_mode
        )



import logging

import packaging.version
import requests

logger = logging.getLogger("bluequbit-python-sdk")


def check_version(version):
    local_version = packaging.version.parse(version)
    if local_version.is_prerelease:
        logger.warning(
            "Beta version %s of BlueQubit Python SDK is being used.", version
        )
    req = requests.get("https://pypi.python.org/pypi/bluequbit/json", timeout=2.0)
    if not req.ok:
        message = "PyPI version check unsuccessful."
        logger.debug(message)
        return message

    # find max version on PyPI
    releases = req.json().get("releases", [])
    pip_version = packaging.version.parse("0")
    for release in releases:
        ver = packaging.version.parse(release)
        if not ver.is_prerelease or local_version.is_prerelease:
            pip_version = max(pip_version, ver)

    message = (
        "There is a %s of BlueQubit Python SDK available on PyPI. We"
        " recommend upgrading. Run 'pip install --upgrade bluequbit' to upgrade"
        " from your version %s to %s."
    )
    if pip_version.major > local_version.major:
        message = message % ("major upgrade", local_version, pip_version)
        logger.warning(message)
    elif pip_version > local_version:
        message = message % ("newer version", local_version, pip_version)
        logger.info(message)
    else:
        message = ""
    return message



from .circuits import (
    QFT,
    DraperAdder,
    bell_pair,
    coin_toss,
    ghz,
    multi_adder,
    random_circuit,
)

__all__ = [
    "QFT",
    "DraperAdder",
    "bell_pair",
    "coin_toss",
    "ghz",
    "multi_adder",
    "random_circuit",
]



from qiskit import QuantumCircuit


def ghz(n: int = 3) -> QuantumCircuit:
    r"""Return circuit that produces the generalized Greenberger-Horne-Zeilinger \(GHZ\) state with n qubits
    defined by:

    .. math::

        |{\rm GHZ} \rangle = \frac{|0 \rangle ^{\otimes n} + |1 \rangle ^{\otimes n}}{\sqrt{2}} , n > 2

    If the argument `n` isn't passed in, the default 3 is used.

    :param n: The number of qubits \(default is 3\). Valid choices are integers greater than 2.
            If you want ``n = 2``, use ``bell_pair`` function from ``bluequbit.library``
    :raises ValueError: if ``n`` is lower than 3
    :return: GHZ circuit
    """
    if n < 3:
        raise ValueError("n must be at least 3")
    ghz_qc = QuantumCircuit(n)
    ghz_qc.h(0)
    ghz_qc.cx(0, range(1, n))
    return ghz_qc



import numpy as np

NUMERIC_TYPES = (int, float, np.int32, np.int64, np.float32, np.float64)


def is_numeric(obj):
    return isinstance(obj, NUMERIC_TYPES)


def parse_pauli_sum_str(psum_str):
    psum_str = psum_str.replace(" ", "")
    # Rename ** so that it doesn't get removed
    psum_str = psum_str.replace("**", "EXPONENT")
    psum_str = psum_str.replace("*", "")
    psum_str = psum_str.replace("EXPONENT", "**")
    if "j" in psum_str:
        raise Exception("Pauli string coefficient must not be complex")
    if "x" in psum_str:
        raise Exception("Use * for multiplication instead of x")
    pstrings = psum_str.split("+")

    def parse_pauli_string(pstring):
        if len(pstring) == 0:
            raise Exception("Empty Pauli string element")
        first_operator_index = 0
        for i, c in enumerate(pstring):
            if c in "IXYZ":
                first_operator_index = i
                break
        if i == 0:
            # No number is specified
            coefficient = 1.0
        elif i == 1 and pstring[0] == "-":
            coefficient = -1.0
        else:
            coefficient = float(pstring[:first_operator_index])
        pstring_itself = pstring[first_operator_index:]
        return pstring_itself, coefficient

    out = [parse_pauli_string(pstring) for pstring in pstrings]

    # Sanity check
    pstring_length = len(out[0][0])
    for ps, _ in out:
        if len(ps) != pstring_length:
            raise Exception(
                "One of the Pauli strings has mismatched length with the first Pauli"
                " string"
            )

    return out


def format_pauli_sum(pauli_sum):
    if not isinstance(pauli_sum, (str, list, tuple)):
        raise TypeError(f"Unsupported format type for Pauli sum: {type(pauli_sum)}")
    if isinstance(pauli_sum, str):
        return [parse_pauli_sum_str(pauli_sum)]
    if len(pauli_sum) == 0:
        raise Exception("The Pauli sum is an empty list/tuple")

    def isa_pauli_string_tuple(e):
        first_condition = len(e) == 2 and isinstance(e[0], str)
        if first_condition and isinstance(e[1], complex):
            raise Exception(
                f"The Pauli string tuple coefficient must not be complex {e}"
            )
        return first_condition and is_numeric(e[1])

    def parse(e):
        if isinstance(e, str):
            return parse_pauli_sum_str(e)
        if isinstance(e, (list, tuple)):
            if isa_pauli_string_tuple(e):
                # pauli_sum is a single Pauli sum
                return e
            if all(isa_pauli_string_tuple(ee) for ee in e):
                # pauli_sum is a list of Pauli sum
                return e
            raise Exception(
                f"Unsupported format type for Pauli sum element {e}: {type(e)}"
            )
        raise Exception(f"Unsupported format type for Pauli sum element {e}: {type(e)}")

    return [parse(e) for e in pauli_sum]



from __future__ import annotations

import contextlib

from . import exceptions, job_metadata_constants
from .bluequbit_client import BQClient
from .bluequbit_logger import init_logger
from .estimate_result import EstimateResult
from .job_result import JobResult

with contextlib.suppress(ImportError):
    from .qiskit_provider.backend import BlueQubitBackendV2
    from .qiskit_provider.estimator import BlueQubitEstimatorV2
    from .qiskit_provider.provider import BlueQubitProvider
    from .qiskit_provider.sampler import BlueQubitSamplerV2

from .version import __version__

with contextlib.suppress(ImportError):
    from .__init__private import *  # noqa: F403

__all__ = [
    "BQClient",
    "BlueQubitBackendV2",
    "BlueQubitEstimatorV2",
    "BlueQubitProvider",
    "BlueQubitSamplerV2",
    "EstimateResult",
    "JobResult",
    "__version__",
    "exceptions",
    "job_metadata_constants",
]

logger = init_logger()


def init(api_token: str | None = None, execution_mode: str | None = None) -> BQClient:
    """Returns :class:`BQClient` instance for managing jobs on BlueQubit platform.

    :param api_token: API token of the user. If ``None``, the token will be looked
                      in the environment variable BLUEQUBIT_API_TOKEN.
    :param execution_mode: Execution mode. Can be ``"cloud"``, ``"local"`` or ``None``.
                           If ``None``, the environment variable
                           BLUEQUBIT_EXECUTION_MODE will be used if set, otherwise
                           ``"cloud"`` will be used.
    """
    return BQClient(api_token, execution_mode)



class EstimateResult:  # pylint: disable=too-few-public-methods
    """This class contains information about the estimated runtime/cost that
    will be incurred from running the given circuit on the given quantum
    machine/simulator. `WARNING:` this is just an estimate, the actual runtime/cost
    may be less or more."""

    def __init__(self, data):
        self.device = data.get("device")

        #: int: estimated runtime in milliseconds
        self.estimated_runtime = data.get("estimate_ms")

        #: float: estimated cost in US dollars
        self.estimated_cost = data.get("estimated_cost")
        if self.estimated_cost is not None:
            self.estimated_cost /= 100.0

        #: int: number of qubits
        self.num_qubits = data.get("num_qubits")

        #: str: warning message
        self.warning_message = data.get("warning_message")

        #: str: error message if available
        self.error_message = data.get("error_message")

    def __repr__(self):
        if self.error_message is not None:
            return f"Estimation failed due to error: {self.error_message}."
        return (
            f"Estimated: runtime: {self.estimated_runtime} ms, cost:"
            f" ${self.estimated_cost:.2f}. {self.warning_message}"
        )



from pathlib import Path

from setuptools import setup


def exclude_lines_between_markers(
    file_path: str,
    start_marker: str = "<!-- EXCLUDE -->",
    end_marker: str = "<!-- /EXCLUDE -->",
) -> str:
    """
    Reads a file and returns its content with lines between the given markers excluded.
    """
    output_lines = []
    exclude = False
    with Path(file_path).open(encoding="utf-8") as f:
        for line in f:
            if start_marker in line and not exclude:
                exclude = True
                continue
            if end_marker in line and exclude:
                exclude = False
                continue
            if not exclude:
                output_lines.append(line)
    return "".join(output_lines)


readme = exclude_lines_between_markers("README.md")

with Path("src/bluequbit/version.py").open() as f:
    Version = f.read()

Version = Version.rstrip()
Version = Version[15:-1]

setup(
    version=Version,
    long_description=readme,
    long_description_content_type="text/markdown",
)



from __future__ import annotations

import io
import logging
import os
import sys
from typing import TYPE_CHECKING, Any

if TYPE_CHECKING:
    from collections.abc import MutableMapping

MSG_PREFIX = " - "


class PrefixLoggerAdapter(logging.LoggerAdapter):
    """A logger adapter that adds a prefix to every message"""

    def process(self, msg: str, kwargs: MutableMapping[str, str]) -> Any:
        msg = MSG_PREFIX + msg
        if self.extra is not None:
            prefix = self.extra["prefix"]
            msg = f"[{prefix}]" + msg
        return msg, kwargs


YELLOW = "\033[33m"
RED = "\033[31m"
CYAN = "\033[36m"
GREEN = "\033[32m"
BLUE = "\033[34m"
RESET = "\033[0m"

BQ_FORMAT = (
    "{color0}[BQ-PYTHON-SDK]{reset}[{color}%(levelname)s{reset}]%(message)s{reset}"
)


class BlueQubitLoggerFormatter(logging.Formatter):
    class Default(dict):
        def __missing__(self, key):
            return "{" + key + "}"

    BQ_COLORED_FORMAT = BQ_FORMAT.format_map(
        Default(
            color0=CYAN,
            reset=RESET,
        )
    )

    FORMATS = {  # noqa: RUF012
        logging.DEBUG: BQ_COLORED_FORMAT.format(color=BLUE),
        logging.INFO: BQ_COLORED_FORMAT.format(color=GREEN),
        logging.WARNING: BQ_COLORED_FORMAT.format(color=YELLOW),
        logging.ERROR: BQ_COLORED_FORMAT.format(color=RED),
        logging.CRITICAL: BQ_COLORED_FORMAT.format(color=RED),
    }

    def __init__(self, *, colored):
        super().__init__()
        self.colored = colored
        self.formatters = {}
        for levelno, log_fmt in self.FORMATS.items():
            if self.colored:
                self.formatters[levelno] = logging.Formatter(log_fmt)
            else:
                self.formatters[levelno] = logging.Formatter(
                    fmt=BQ_FORMAT.format(color0="", reset="", color="")
                )

    def format(self, record):
        return self.formatters[record.levelno].format(record)


# modified from https://github.com/termcolor/termcolor/blob/main/src/termcolor/termcolor.py
def _can_do_colour() -> bool:
    """Check env vars and for tty/dumb terminal"""
    # Then check env vars:
    if "ANSI_COLORS_DISABLED" in os.environ:
        return False
    if "NO_COLOR" in os.environ:
        return False
    if "FORCE_COLOR" in os.environ:
        return True

    # Then check system:
    if os.environ.get("TERM") == "dumb":
        return False
    if not hasattr(sys.stdout, "fileno"):
        return False
    if not hasattr(sys.stderr, "fileno"):
        return False

    try:
        return os.isatty(sys.stdout.fileno()) and os.isatty(sys.stderr.fileno())
    except io.UnsupportedOperation:
        return sys.stdout.isatty() and sys.stderr.isatty()


class InfoFilter(logging.Filter):
    def filter(self, record):
        return record.levelno in (logging.DEBUG, logging.INFO)


class PrefixWithSpaceFilter(logging.Filter):
    def filter(self, record):
        if record.msg[0] != "[":
            record.msg = MSG_PREFIX + record.msg
        return True


def init_logger():
    handler1 = logging.StreamHandler(sys.stdout)
    handler2 = logging.StreamHandler(sys.stderr)
    if _can_do_colour():
        handler1.setFormatter(BlueQubitLoggerFormatter(colored=True))
        handler2.setFormatter(BlueQubitLoggerFormatter(colored=True))
    else:
        handler1.setFormatter(BlueQubitLoggerFormatter(colored=False))
        handler2.setFormatter(BlueQubitLoggerFormatter(colored=False))

    handler1.setLevel(logging.DEBUG)
    handler1.addFilter(InfoFilter())

    handler2.setLevel(logging.WARNING)

    handler1.addFilter(PrefixWithSpaceFilter())
    handler2.addFilter(PrefixWithSpaceFilter())

    logger = logging.getLogger("bluequbit-python-sdk")
    logger.setLevel(logging.INFO)
    logger.addHandler(handler1)
    logger.addHandler(handler2)
    return logger



from .bell_pair import bell_pair  # noqa: F401
from .coin_toss import coin_toss  # noqa: F401
from .draper_adder import DraperAdder  # noqa: F401
from .ghz import ghz  # noqa: F401
from .multi_adder import multi_adder  # noqa: F401
from .qft import QFT  # noqa: F401
from .random_circuit import random_circuit  # noqa: F401



from .bluequbit_device import BluequbitDevice


class BluequbitCPU(BluequbitDevice):
    """BluequbitCPU device for PennyLane. This device is used to run Pennylane circuits on BlueQubit platform using a
    cpu device and is similar to using lightning.qubit on your local device. It requires a BlueQubit token, which
    you can get from your BlueQubit account.

    .. warning::

            To use this plugin, you must have pennylane>=0.39 version installed. It requires
            Python 3.9, but we would recommend using Python 3.10 . Make sure your Python version
            is not older.

    Args:
        wires (int, Iterable[Number, str]): Number of subsystems represented by the device,
            or iterable that contains unique labels for the subsystems as numbers (i.e., ``[-1, 0, 2]``)
            or strings (``['auxiliary', 'q1', 'q2']``). Default 1 if not specified.
        token (str, None): Your BlueQubit token. This is a required keyword-only argument, so you have to specify
            it every time, even when you give your token through the environment variable. If you have set your token
            in BLUEQUBIT_API_TOKEN environment variable you should pass token=None to the device.
    """

    short_name = "bluequbit.cpu"

    def __init__(self, wires, *args, token, **kwargs):
        super().__init__(wires, *args, token=token, **kwargs)
        self._device = "pennylane.cpu"



from qiskit import QuantumCircuit


def coin_toss(n: int = 1) -> QuantumCircuit:
    r"""Return circuit that performs n coin tosses
    If the argument `n` isn't passed in, the default 1 is used.

    :param n: The number of coin tosses \(default is 1\). Valid choices are positive integers
    :raises ValueError: If ``n`` is lower than 1
    :return: Circuit performing ``n`` coin tosses
    """
    if n < 1:
        raise ValueError("n must be a positive integer")
    coin_toss_qc = QuantumCircuit(n)
    coin_toss_qc.h(range(n))
    return coin_toss_qc
